﻿using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Generators.CodeGenerator {

  using HeuristicLab.Core;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc.Interfaces;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  internal static class CodeGeneratorUtils {
    internal static Expression CreateExpressionOrErc(
      IRandom random,
      IReadOnlyList<string> enabledExpressions,
      IReadOnlyErcOptions ercOptions,
      IDictionary<string, Expression> customExpressions = null) {
      var customCount = customExpressions == null ? 0 : customExpressions.Count;
      var index = random.Next(enabledExpressions.Count + customCount);

      return index >= enabledExpressions.Count
        ? customExpressions.ElementAt(index - (enabledExpressions.Count - 1)).Value
        : CreateExpressionOrErc(index, random, enabledExpressions, ercOptions);
    }

    internal static Expression CreateExpressionOrErc(
      int index,
      IRandom random,
      IReadOnlyList<string> enabledExpressions,
      IReadOnlyErcOptions ercOptions) {
      var name = enabledExpressions[index];

      return CreateExpressionOrErc(name, random, ercOptions);
    }

    private static Expression CreateExpressionOrErc(
      string expressionName,
      IRandom random,
      IReadOnlyErcOptions ercOptions) {
      var x = random.NextDouble();
      Expression expression = null;

      if (x < ercOptions.ErcProbability) {
        var expressionType = ExpressionTable.NameToTypeTable[expressionName];
        expression = CreateRandomErcExpression(
            ExpressionTable.TypeToAttributeTable[expressionType].StackType,
            random,
            ercOptions);
      }

      if (expression == null || expression is ExecNoopExpression) {
        expression = ExpressionTable.GetExpression(expressionName);
      }

      return expression;
    }

    /// <summary>
    /// Create a ErcExpression whereby the type of the expression conforms to the passed stack type.
    /// </summary>
    /// <param name="type"></param>
    /// <param name="random"></param>
    /// <param name="pushConfiguration"></param>
    /// <returns>ErcExpression or Noop if the required ErcOptionConstants is not configured or enabled in the passed config.</returns>
    internal static Expression CreateRandomErcExpression(StackTypes type, IRandom random, IReadOnlyErcOptions ercOptions) {
      switch (type) {
        case StackTypes.Integer:
          if (ercOptions.IntegerErcOptions == null || !ercOptions.IntegerErcOptions.IsEnabled)
            return ExpressionTable.GetStatelessExpression<ExecNoopExpression>();

          var intValue = ErcUtils.GetInteger(ercOptions.IntegerErcOptions, random);

          return new IntegerPushExpression(intValue);

        case StackTypes.Float:
          if (ercOptions.FloatErcOptions == null || !ercOptions.FloatErcOptions.IsEnabled)
            return ExpressionTable.GetStatelessExpression<ExecNoopExpression>();

          var doubleValue = ErcUtils.GetDouble(ercOptions.FloatErcOptions, random);

          return new FloatPushExpression(doubleValue);

        case StackTypes.Boolean:
          return ercOptions.BooleanErcOptions == null || !ercOptions.BooleanErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new BooleanPushExpression(ErcUtils.GetBoolean(ercOptions.BooleanErcOptions, random));

        case StackTypes.Name:
          return ercOptions.NameErcOptions == null || !ercOptions.NameErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new NamePushExpression(ErcUtils.GetName(ercOptions.NameErcOptions, random));

        case StackTypes.String:
          return ercOptions.StringErcOptions == null || !ercOptions.StringErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new StringPushExpression(ErcUtils.GetString(ercOptions.StringErcOptions, random));

        case StackTypes.Char:
          return ercOptions.CharErcOptions == null || !ercOptions.CharErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new CharPushExpression(ErcUtils.GetChar(ercOptions.CharErcOptions, random));

        default:
          return ExpressionTable.GetStatelessExpression<ExecNoopExpression>();
      }
    }
  }
}
