﻿using System;
using System.Collections.Generic;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {

  using HeuristicLab.BenchmarkSuite;
  using HeuristicLab.BenchmarkSuite.Problems;
  using HeuristicLab.Common;
  using HeuristicLab.Core;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  public static class PushEvaluator {
    public static double Evaluate(IPushInterpreter interpreter, PushProgram program, Example example, double worstResult) {
      interpreter.BooleanStack.Push(example.InputBoolean);
      interpreter.IntegerStack.Push(example.InputInteger);
      interpreter.FloatStack.Push(example.InputFloat);

      interpreter.Run(program);

      var result = GetDiff(example.OutputInteger, interpreter.IntegerStack, worstResult, IntegerDiffer)
                 + GetDiff(example.OutputFloat, interpreter.FloatStack, worstResult, FloatDiffer)
                 + GetDiff(example.OutputBoolean, interpreter.BooleanStack, worstResult, BooleanDiffer)
                 + GetDiff(example.OutputString, interpreter.StringStack, worstResult, StringDiffer)
                 + GetDiff(example.OutputChar, interpreter.CharStack, worstResult, CharDiffer)
                 + GetVectorDiff(example.OutputIntegerVector, interpreter.IntegerVectorStack, worstResult, (a, b) => VectorDiffer(a, b, IntegerDiffer))
                 + GetVectorDiff(example.OutputFloatVector, interpreter.FloatVectorStack, worstResult, (a, b) => VectorDiffer(a, b, FloatDiffer))
                 + GetVectorDiff(example.OutputStringVector, interpreter.StringVectorStack, worstResult, (a, b) => VectorDiffer(a, b, StringDiffer));

      return result;
    }

    public static EvaluationResult Evaluate(PushProgram program, PushInterpreterPool pool, IRandom random, ProblemData data, int startIndex, int endIndex) {
      var length = endIndex - startIndex;
      if (length <= 0) return null;

      var evaluationResult = new EvaluationResult(length);

      using (var interpreter = pool.Create(random)) {
        for (var i = startIndex; i < endIndex; i++) {
          var result = Evaluate(interpreter, program, data.Examples[i], data.WorstResult);
          evaluationResult.ExampleQualities[i - startIndex] = result;
          evaluationResult.TotalQuality += result;
          interpreter.Reset();
        }
      }

      evaluationResult.TotalQuality /= length;

      return evaluationResult;
    }

    private static double FloatDiffer(double a, double b) {
      var result = a - b;

      if (result.IsAlmost(double.MinValue) || double.IsPositiveInfinity(result) || double.IsNaN(result))
        return double.MaxValue;

      if (result.IsAlmost(double.MaxValue) || double.IsNegativeInfinity(result))
        return double.MinValue;

      return Math.Abs(result);
    }

    private static double IntegerDiffer(long a, long b) {
      var result = a - b;

      return result == long.MinValue ? long.MaxValue : Math.Abs(result);
    }

    private static double BooleanDiffer(bool a, bool b) {
      return a && b ? 0 : a || b ? 1 : 2;
    }

    private static double StringDiffer(string a, string b) {
      return LevenshteinDistance(a, b);
    }

    private static double CharDiffer(char a, char b) {
      return IntegerDiffer(a, b);
    }

    private static double VectorDiffer<T>(IReadOnlyList<T> a, IReadOnlyList<T> b, Func<T, T, double> differ) {
      var length = Math.Min(a.Count, b.Count);
      var result = 0d;

      for (var i = 0; i < length; i++) {
        result += differ(a[i], b[i]);
      }

      if (a.Count > b.Count) {
        var remainingLength = a.Count - length;
        for (var i = 0; i < remainingLength; i++)
          result += differ(a[i], default(T));
      }

      return result;
    }

    private static double GetVectorDiff<T>(IReadOnlyList<IReadOnlyList<T>> estimated, IPushStack<List<T>> resultStack, double worstResult, Func<IReadOnlyList<T>, IReadOnlyList<T>, double> differ)
      where T : IComparable {
      if (estimated.Count == 0) return 0d;

      var diff = 0d;
      var comparableLength = 0;

      if (!resultStack.IsEmpty) {
        var count = Math.Min(estimated.Count, resultStack.Count);
        var result = resultStack.Peek(count);
        comparableLength = Math.Min(estimated.Count, result.Length);

        for (var i = 0; i < comparableLength; i++) {
          diff += Math.Min(differ(estimated[i], result[i]), worstResult);
        }
      }

      var emptyTArray = new T[0];
      for (var i = comparableLength; i < estimated.Count - comparableLength; i++) {
        diff += differ(estimated[i], emptyTArray);
      }

      return diff;
    }

    private static double GetDiff<T>(IReadOnlyList<T> estimated, IPushStack<T> resultStack, double worstResult, Func<T, T, double> differ)
      where T : IComparable {
      if (estimated.Count == 0) return 0d;

      var diff = 0d;
      var comparableLength = 0;

      if (!resultStack.IsEmpty) {
        var count = Math.Min(estimated.Count, resultStack.Count);
        var result = resultStack.Peek(count);
        comparableLength = Math.Min(estimated.Count, result.Length);

        for (var i = 0; i < comparableLength; i++) {
          diff += Math.Min(differ(estimated[i], result[i]), worstResult);
        }
      }

      for (var i = comparableLength; i < estimated.Count - comparableLength; i++) {
        diff += differ(estimated[i], default(T));
      }

      return diff;
    }

    /// <summary>
    /// https://www.dotnetperls.com/levenshtein
    /// </summary>
    /// <param name="s"></param>
    /// <param name="t"></param>
    /// <returns></returns>
    private static int LevenshteinDistance(string s, string t) {
      if (s == null && t == null) return 0;
      if (s == null) return t.Length;
      if (t == null) return s.Length;

      int n = s.Length;
      int m = t.Length;
      int[,] d = new int[n + 1, m + 1];

      // Step 1
      if (n == 0) {
        return m;
      }

      if (m == 0) {
        return n;
      }

      // Step 2
      for (int i = 0; i <= n; d[i, 0] = i++) {
      }

      for (int j = 0; j <= m; d[0, j] = j++) {
      }

      // Step 3
      for (int i = 1; i <= n; i++) {
        //Step 4
        for (int j = 1; j <= m; j++) {
          // Step 5
          int cost = (t[j - 1] == s[i - 1]) ? 0 : 1;

          // Step 6
          d[i, j] = Math.Min(
              Math.Min(d[i - 1, j] + 1, d[i, j - 1] + 1),
              d[i - 1, j - 1] + cost);
        }
      }
      // Step 7
      return d[n, m];
    }
  }
}
